import csv
import shutil
from datetime import datetime
from os.path import dirname, join
from typing import Dict, List, Set, Tuple

import pandas as pd
from fideslang.models import DataSubject, DataSubjectRightsEnum, DataUse
from fideslang.validation import FidesKey

DATAMAP_TEMPLATE = join(
    dirname(__file__),
    "../templates",
    "fides_datamap_template.xlsx",
)

DATAMAP_COLUMNS = {
    "dataset.name": "Fides Dataset",
    "system.name": "Fides System",
    "system.administrating_department": "Department or Business Function",
    "system.privacy_declaration.data_use.name": "Purpose of Processing",
    "system.joint_controller": "Joint Controller",
    "system.privacy_declaration.data_subjects.name": "Categories of Individuals",
    "unioned_data_categories": "Categories of Personal Data (Fides Taxonomy)",
    "system.privacy_declaration.data_use.recipients": "Categories of Recipients",
    "system.link_to_processor_contract": "Link to Contract with Processor",
    "third_country_combined": "Third Country Transfers",
    "system.third_country_safeguards": "Safeguards for Exceptional Transfers of Personal Data",
    "dataset.retention": "Retention Schedule",
    "organization.link_to_security_policy": "General Description of Security Measures",
    "system.data_responsibility_title": "Role or Responsibility",
    "system.privacy_declaration.data_use.legal_basis": "Article 6 lawful basis for processing personal data",
    "system.privacy_declaration.data_use.special_category": "Article 9 condition for processing special category data",
    "system.privacy_declaration.data_use.legitimate_interest": "Legitimate interests for the processing (if applicable)",
    "system.privacy_declaration.data_use.legitimate_interest_impact_assessment": "Link to record of legitimate interests assessment (if applicable)",
    "system.privacy_declaration.data_subjects.rights_available": "Rights available to individuals",
    "system.privacy_declaration.data_subjects.automated_decisions_or_profiling": "Existence of automated decision-making, including profiling (if applicable)",
    "dataset.source_name": "The source of the personal data (if applicable)",
    "system.data_protection_impact_assessment.is_required": "Data Protection Impact Assessment required?",
    "system.data_protection_impact_assessment.progress": "Data Protection Impact Assessment progress",
    "system.data_protection_impact_assessment.link": "Link to Data Protection Impact Assessment",
}


def export_to_csv(
    list_to_export: List, resource_exported: str, manifests_dir: str
) -> str:
    """
    Exports a list of Tuples of any length back to the manifest
    directory as a csv.
    """
    utc_datetime = datetime.utcnow().strftime("%Y-%m-%d-T%H%M%S")
    filename = f"{utc_datetime}_{resource_exported}.csv"
    filepath = f"{manifests_dir}/{filename}"
    with open(filepath, "w", encoding="utf-8") as csvfile:
        writer = csv.writer(csvfile)
        writer.writerows(list_to_export)

    return filename


def generate_data_category_rows(
    dataset_name: str,
    dataset_description: str,
    data_qualifier: str,
    data_categories: List,
    data_retention: str,
    dataset_third_country_transfers: str,
    dataset_fides_key: str,
) -> Set[Tuple[str, str, str, str, str, str, str]]:
    """
    Set comprehension to capture categories from any
    of the possible levels in a dataset resource.

    Returns a set of tuples to be unioned with the
    overall set to be exported.
    """
    dataset_rows = {
        (
            dataset_name,
            dataset_description,
            category,
            data_qualifier,
            data_retention,
            dataset_third_country_transfers,
            dataset_fides_key,
        )
        for category in data_categories
    }

    return dataset_rows


def export_datamap_to_excel(
    organization_df: pd.DataFrame,
    joined_system_dataset_df: pd.DataFrame,
    manifests_dir: str,
) -> str:
    """
    Creates a copy from the existing datamap template and generates
    an exported datamap based on applied fides resources. Uses two
    dataframes, one for the organization and one for the joined
    system and dataset resources.
    """
    utc_datetime = datetime.utcnow().strftime("%Y-%m-%d-T%H%M%S")
    filename = f"{utc_datetime}_datamap.xlsx"
    filepath = f"{manifests_dir}/{filename}"

    shutil.copy(DATAMAP_TEMPLATE, filepath)

    # pylint: disable=abstract-class-instantiated
    with pd.ExcelWriter(
        filepath, mode="a", if_sheet_exists="overlay"
    ) as export_file:  # pylint: enable=abstract-class-instantiated
        organization_df.to_excel(
            export_file,
            sheet_name="sheet1",
            index=False,
            header=False,
            startrow=2,
            startcol=2,
        )

        joined_system_dataset_df.sort_values(
            by=[
                "system.name",
                "unioned_data_categories",
            ]
        ).to_excel(
            export_file,
            sheet_name="sheet1",
            index=False,
            header=False,
            startrow=9,
            columns=list(DATAMAP_COLUMNS.keys()),
        )

    return filename


def format_data_uses(data_uses: List[DataUse]) -> Dict[FidesKey, Dict[str, str]]:
    """
    This function formats data uses for use when exporting,
    returning the necessary values as a dict. Formatting
    differences exist due to various types allowed across attributes.
    """

    formatted_data_uses = {}
    for data_use in data_uses:
        formatted_data_use = {
            "name": data_use.name,
        }

        for attribute in [
            "legal_basis",
            "special_category",
            "recipients",
            "legitimate_interest_impact_assessment",
            "legitimate_interest",
        ]:
            attribute_value = getattr(data_use, attribute)
            if attribute_value is None:
                attribute_value = "N/A"
            elif isinstance(attribute_value, list):
                attribute_value = ", ".join(attribute_value)
            elif attribute == "legitimate_interest":
                if attribute_value is True:
                    attribute_value = getattr(data_use, "name")
                else:
                    attribute_value = "N/A"

            formatted_data_use[attribute] = attribute_value

        formatted_data_uses[data_use.fides_key] = formatted_data_use
    return formatted_data_uses


def format_data_subjects(
    data_subjects: List[DataSubject],
) -> Dict[FidesKey, Dict[str, str]]:
    """
    This function formats data subjects from the server, returning the necessary values
    as a list of dicts.

    rights_available is treated differently due to the
    different strategy options available in returning
    the available data subject rights.
    """

    formatted_data_subject_attributes_list = [
        "name",
        "rights_available",
        "automated_decisions_or_profiling",
    ]

    formatted_data_subjects: Dict[FidesKey, Dict[str, str]] = {}

    for data_subject in data_subjects:
        data_subject_dict = data_subject.dict()
        formatted_data_subject = dict(
            zip(
                formatted_data_subject_attributes_list,
                [
                    "N/A",
                ]
                * len(formatted_data_subject_attributes_list),
            )
        )

        # calculate and format data subject rights as applicable
        if data_subject_dict["rights"]:
            data_subject_dict["rights_available"] = calculate_data_subject_rights(
                data_subject_dict["rights"]
            )
        else:
            data_subject_dict["rights_available"] = "No data subject rights listed"

        formatted_data_subject = {
            attribute: data_subject_dict.get(attribute) or "N/A"
            for attribute in formatted_data_subject_attributes_list
        }

        formatted_data_subjects[data_subject.fides_key] = formatted_data_subject

    return formatted_data_subjects


def convert_tuple_to_string(values: Tuple[str, ...]) -> str:
    """
    Takes a tuple of variable length strings and combines them into a comma seperated
    string.

    During the conversion empty strings, "", are converted to "N/A", duplicates are removed,
    then if is checked to see if all values in the tuple are "N/A" if them are the string
    "N/A" is returned. If there are multiple strings presents and some are "N/A", the
    "N/A" is dropped in the resulting string.

    Example:
        - ("CAN", "GBR", "USA") returns "CAN, GBR, USA"
        - ("CAN, GBR", "CAN", "GBR") returns "CAN GBR"
        - ("N/A", "CAN", "GBR") returns "CAN, GBR"
        - ("", "CAN") returns "CAN"
        - ("", "N/A") returns "N/A"
        - ("N/A", "N/A") returns "N/A"
    """
    empty_replaced = {"N/A" if x == "" else x for x in values}
    if empty_replaced == {"N/A"}:
        return "N/A"

    return remove_duplicates_from_comma_separated_column(
        ", ".join((x for x in empty_replaced if x != "N/A"))
    )


def calculate_data_subject_rights(rights: Dict) -> str:
    """
    Calculate and format the data subject individual rights.

    Loads all available rights
    """
    all_rights = DataSubjectRightsEnum
    strategy: str = rights["strategy"]
    data_subject_rights: str
    if strategy == "ALL":
        data_subject_rights = ", ".join(all_rights)
    elif strategy == "INCLUDE":
        included_rights = rights["values"]
        data_subject_rights = ", ".join(included_rights)
    elif strategy == "EXCLUDE":
        excluded_rights = rights["values"]
        included_rights = [
            right for right in all_rights if right not in excluded_rights
        ]
        data_subject_rights = ", ".join(included_rights)
    elif strategy == "NONE":
        data_subject_rights = "No data subject rights listed"

    return data_subject_rights


def remove_duplicates_from_comma_separated_column(comma_separated_string: str) -> str:
    "transform the row using a set to remove duplcation"
    return ", ".join(set(comma_separated_string.split(", ")))


def union_data_categories_in_joined_dataframe(joined_df: pd.DataFrame) -> pd.DataFrame:
    """
    Data Categories can be present in both the System and (optionally) Dataset
    resources, causing duplication when joining them together.

    This function isolates the data categories from each into a new column which
    when unioned together can be used to populate the data map accurately with
    the data categories from both System and Dataset.
    """

    # isolate the system data categories into a new dataframe and create a common column
    joined_df = joined_df.drop(
        [
            "system.privacy_declaration.name",
            "dataset.description",
        ],
        axis=1,
    )
    systems_categories_df = joined_df.drop(
        [
            "dataset.data_categories",
            "dataset.retention",
            "dataset.data_qualifier",
        ],
        axis=1,
    )
    systems_categories_df["dataset.name"] = "N/A"
    systems_categories_df["unioned_data_categories"] = systems_categories_df[
        "system.privacy_declaration.data_categories"
    ]
    systems_categories_df["unioned_data_qualifiers"] = systems_categories_df[
        "system.privacy_declaration.data_qualifier"
    ]
    systems_categories_df["dataset.retention"] = "N/A"
    systems_categories_df["dataset.data_qualifier"] = "N/A"

    # isolate the dataset data categories into a new dataframe and create a common column
    datasets_categories_df = joined_df.drop(
        [
            "system.privacy_declaration.data_categories",
            "system.privacy_declaration.data_qualifier",
        ],
        axis=1,
    )

    datasets_categories_df = datasets_categories_df[
        datasets_categories_df["dataset.name"] != "N/A"
    ]

    datasets_categories_df["unioned_data_qualifiers"] = datasets_categories_df[
        "dataset.data_qualifier"
    ]
    datasets_categories_df["unioned_data_categories"] = datasets_categories_df[
        "dataset.data_categories"
    ]

    # union the two dataframes together to return all data categories
    return pd.concat(
        [
            systems_categories_df.drop(
                ["system.privacy_declaration.data_categories"], axis="columns"
            ),
            datasets_categories_df.drop(["dataset.data_categories"], axis="columns"),
        ]
    ).drop_duplicates()


def get_formatted_data_protection_impact_assessment(
    data_protection_impact_assessment: dict,
) -> dict:
    "Replace None with N/A for consistent formatting of the data map"
    return {
        key: ("N/A" if value is None else value)
        for key, value in data_protection_impact_assessment.items()
    }



from __future__ import annotations

import json
from typing import Any

from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy_utils import JSONType


class JSONTypeOverride(JSONType):  # pylint: disable=W0223
    """
    Temporary override of JSONType with workaround to fix the return type.

    Using a StringEncryptedType with JSONType as the underlying type returns a string instead of a dictionary:
    '{"key": "value"}' instead of {"key": "value"}

    https://github.com/kvesteri/sqlalchemy-utils/issues/532
    """

    def process_bind_param(self, value: MutableDict, _: Any | None) -> str | None:
        """
        Overrides JSONType.process_bind_param to return json.dumps(value) instead of just the value.
        """
        if value is not None:
            return json.dumps(value)
        return value

    def process_result_value(self, value: str, _: Any | None) -> dict[str, Any] | None:
        """
        Overrides JSONType.process_result_value to return json.loads(value) instead of just the value.
        """
        if value is not None:
            return json.loads(value)
        return value



"""unified fides 2 merger

Revision ID: cdbd0fa118dd
Revises: 6f98e1bd7e4f, cf88efa1ad89
Create Date: 2022-09-28 16:27:25.429064

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "cdbd0fa118dd"
down_revision = ("6f98e1bd7e4f", "cf88efa1ad89")
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



from abc import abstractmethod

from requests import PreparedRequest

from fides.api.ops.models.connectionconfig import ConnectionConfig
from fides.api.ops.service.strategy import Strategy


class AuthenticationStrategy(Strategy):
    """Abstract base class for SaaS authentication strategies"""

    @abstractmethod
    def add_authentication(
        self, request: PreparedRequest, connection_config: ConnectionConfig
    ) -> PreparedRequest:
        """Add authentication to the request"""



"""adds first_name and last_name to user model

Revision ID: 29a7d707163a
Revises: 90070db16d05
Create Date: 2022-05-05 13:41:28.807920

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "29a7d707163a"
down_revision = "90070db16d05"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("fidesopsuser", sa.Column("first_name", sa.String(), nullable=True))
    op.add_column("fidesopsuser", sa.Column("last_name", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("fidesopsuser", "last_name")
    op.drop_column("fidesopsuser", "first_name")
    # ### end Alembic commands ###



from __future__ import annotations

from sqlalchemy.orm import Session

from fides.api.ops.models.privacy_request import ConsentRequest, PrivacyRequest
from fides.api.ops.schemas.messaging.messaging import (
    MessagingActionType,
    SubjectIdentityVerificationBodyParams,
)
from fides.api.ops.schemas.redis_cache import Identity
from fides.api.ops.service.messaging.message_dispatch_service import dispatch_message
from fides.api.ops.service.privacy_request.request_runner_service import (
    generate_id_verification_code,
)
from fides.core.config import get_config

CONFIG = get_config()


def send_verification_code_to_user(
    db: Session, request: ConsentRequest | PrivacyRequest, to_identity: Identity | None
) -> str:
    """Generate and cache a verification code, and then message the user"""
    verification_code = generate_id_verification_code()
    request.cache_identity_verification_code(verification_code)
    messaging_action_type = (
        MessagingActionType.CONSENT_REQUEST
        if isinstance(request, ConsentRequest)
        else MessagingActionType.SUBJECT_IDENTITY_VERIFICATION
    )
    dispatch_message(
        db,
        action_type=messaging_action_type,
        to_identity=to_identity,
        service_type=CONFIG.notifications.notification_service_type,
        message_body_params=SubjectIdentityVerificationBodyParams(
            verification_code=verification_code,
            verification_code_ttl_seconds=CONFIG.redis.identity_verification_code_ttl_seconds,
        ),
    )

    return verification_code



from urllib.parse import parse_qs, urlencode, urlsplit, urlunsplit


def set_query_parameter(url: str, param_name: str, param_value: str) -> str:
    """Given a URL, set or replace a query parameter and return the
    modified URL.

    'http://example.com?foo=stuff&biz=baz'
    """

    scheme, netloc, path, query_string, fragment = urlsplit(url)
    query_params = parse_qs(query_string)

    query_params[param_name] = [param_value]
    new_query_string = urlencode(query_params, doseq=True)

    return urlunsplit((scheme, netloc, path, new_query_string, fragment))



"""Module for evaluating policies."""
import uuid
from typing import Callable, Dict, List, Optional, Set, cast

from fideslang.default_taxonomy import DEFAULT_TAXONOMY
from fideslang.models import (
    Dataset,
    Evaluation,
    MatchesEnum,
    Policy,
    PolicyRule,
    PrivacyDeclaration,
    StatusEnum,
    System,
    Taxonomy,
    Violation,
    ViolationAttributes,
)
from fideslang.relationships import get_referenced_missing_keys
from fideslang.utils import get_resource_by_fides_key
from fideslang.validation import FidesKey
from pydantic import AnyHttpUrl

from fides.cli.utils import handle_cli_response, pretty_echo
from fides.core import api
from fides.core.api_helpers import get_server_resource, get_server_resources
from fides.core.parse import parse
from fides.core.utils import echo_green, echo_red, get_all_level_fields


def get_evaluation_policies(
    local_policies: List[Policy],
    evaluate_fides_key: Optional[str],
    url: AnyHttpUrl,
    headers: Dict[str, str],
) -> List[Policy]:
    """
    Returns policies to be evaluated. If 'evaluate_fides_key' is
    passed then only that policy will be returned. Otherwise, returns
    all policies.
    """
    if evaluate_fides_key:
        local_policy_found = next(
            filter(
                lambda policy: policy.fides_key == evaluate_fides_key, local_policies
            ),
            None,
        )
        if local_policy_found:
            return [local_policy_found]

        server_policy_found = get_server_resource(
            url=url,
            resource_type="policy",
            resource_key=evaluate_fides_key,
            headers=headers,
        )
        return [server_policy_found] if server_policy_found else []

    local_policy_keys = (
        [policy.fides_key for policy in local_policies] if local_policies else None
    )
    all_policies = local_policies + get_all_server_policies(
        url=url, headers=headers, exclude=local_policy_keys
    )
    return all_policies


def get_all_server_policies(
    url: AnyHttpUrl, headers: Dict[str, str], exclude: Optional[List[FidesKey]] = None
) -> List[Policy]:
    """
    Get a list of all of the Policies that exist on the server.

    If 'exclude' is passed those specific Policies won't be pulled from the server.
    """

    exclude = exclude if exclude else []
    ls_response = handle_cli_response(
        api.ls(url=url, resource_type="policy", headers=headers), verbose=False
    )
    policy_keys = [
        resource["fides_key"]
        for resource in ls_response.json()
        if resource["fides_key"] not in exclude
    ]
    policy_list = get_server_resources(
        url=url, resource_type="policy", headers=headers, existing_keys=policy_keys
    )
    return policy_list


def validate_policies_exist(policies: List[Policy], evaluate_fides_key: str) -> None:
    """
    Validates that policies to be evaluated exist. If no policies were found
    raises an error and logs an error.
    """
    if not policies:
        echo_red(
            "Policy ({}) could not be found".format(evaluate_fides_key)
            if evaluate_fides_key
            else "No Policies found to evaluate"
        )
        raise SystemExit(1)


def get_fides_key_parent_hierarchy(
    taxonomy: Taxonomy, fides_key: str
) -> List[FidesKey]:
    """
    Traverses a hierarchy of parents for a given fides key and returns
    the hierarchy starting with the given fides key.
    """
    current_key = fides_key
    fides_key_parent_hierarchy = []
    while True:
        fides_key_parent_hierarchy.append(current_key)
        found_resource_map = get_resource_by_fides_key(
            taxonomy=taxonomy, fides_key=current_key
        )
        if found_resource_map:
            found_resource = list(found_resource_map.values())[-1]
            if found_resource and "parent_key" in found_resource.__fields_set__:
                current_key = getattr(found_resource, "parent_key")
                if not current_key:
                    break
            else:
                break
        else:
            echo_red(
                "Found missing key ({}) referenced in taxonomy".format(current_key)
            )
            raise SystemExit(1)
    return fides_key_parent_hierarchy


def compare_rule_to_declaration(
    rule_types: List[FidesKey],
    declaration_type_hierarchies: List[List[FidesKey]],
    rule_match: MatchesEnum,
) -> Set[str]:
    """
    Compare the list of fides_keys within the rule against the list
    of fides_keys hierarchies from the declaration and uses the rule's matches
    field to determine whether the rule is triggered or not. Returns the offending
    keys, prioritizing the first descendant in the hierarchy.
    """
    matched_declaration_types = set()
    mismatched_declaration_types = set()
    for declaration_type_hierarchy in declaration_type_hierarchies:
        declared_declaration_type = declaration_type_hierarchy[0]
        if len(set(declaration_type_hierarchy).intersection(set(rule_types))) > 0:
            matched_declaration_types.add(declared_declaration_type)
        else:
            mismatched_declaration_types.add(declared_declaration_type)

    matches_map: Dict[MatchesEnum, Callable] = {
        # any matches return matching declared values as violations
        MatchesEnum.ANY: lambda: matched_declaration_types,
        # all matches return matching declared values as violations if all values match rule values
        MatchesEnum.ALL: lambda: matched_declaration_types
        if len(matched_declaration_types) == len(declaration_type_hierarchies)
        else set(),
        # none matches return mismatched declared values as violations if none of the values matched rule values
        MatchesEnum.NONE: lambda: mismatched_declaration_types
        if not any(matched_declaration_types)
        else set(),
        # other matches return mismatched declared values as violations
        MatchesEnum.OTHER: lambda: mismatched_declaration_types,
    }

    return matches_map[rule_match]()


def evaluate_policy_rule(
    taxonomy: Taxonomy,
    policy_rule: PolicyRule,
    data_subjects: List[str],
    data_categories: List[str],
    data_qualifier: str,
    data_use: str,
    declaration_violation_message: str,
) -> List[Violation]:
    """
    Given data subjects, data categories, data qualifier and data use,
    builds hierarchies of applicable types and evaluates the result of a
    policy rule
    """
    category_hierarchies = [
        get_fides_key_parent_hierarchy(
            taxonomy=taxonomy, fides_key=declaration_category
        )
        for declaration_category in data_categories
    ]
    data_category_violations = compare_rule_to_declaration(
        rule_types=policy_rule.data_categories.values,
        declaration_type_hierarchies=category_hierarchies,
        rule_match=policy_rule.data_categories.matches,
    )

    # A declaration only has one data use, so its hierarchy gets put in a list
    data_use_hierarchies = [
        get_fides_key_parent_hierarchy(taxonomy=taxonomy, fides_key=data_use)
    ]
    data_use_violations = compare_rule_to_declaration(
        rule_types=policy_rule.data_uses.values,
        declaration_type_hierarchies=data_use_hierarchies,
        rule_match=policy_rule.data_uses.matches,
    )

    # A data subject does not have a hierarchical structure
    data_subject_violations = compare_rule_to_declaration(
        rule_types=policy_rule.data_subjects.values,
        declaration_type_hierarchies=[[data_subject] for data_subject in data_subjects],
        rule_match=policy_rule.data_subjects.matches,
    )

    data_qualifier_violation = (
        policy_rule.data_qualifier
        in get_fides_key_parent_hierarchy(taxonomy=taxonomy, fides_key=data_qualifier)
    )

    evaluation_result = all(
        [
            data_category_violations,
            data_use_violations,
            data_subject_violations,
            data_qualifier_violation,
        ]
    )

    if evaluation_result:
        violations = [
            Violation(
                detail="{}. Violated usage of data categories ({}) with qualifier ({}) for data uses ({}) and subjects ({})".format(
                    declaration_violation_message,
                    ",".join(data_category_violations),
                    data_qualifier,
                    ",".join(data_use_violations),
                    ",".join(data_subject_violations),
                ),
                violating_attributes=ViolationAttributes(
                    data_categories=data_category_violations,
                    data_uses=data_use_violations,
                    data_subjects=data_subject_violations,
                    data_qualifier=data_qualifier,
                ),
            )
        ]
        return violations
    return []


def get_dataset_by_fides_key(taxonomy: Taxonomy, fides_key: str) -> Optional[Dataset]:
    """
    Returns a dataset within the taxonomy for a given fides key
    """
    dataset = next(
        iter(
            [dataset for dataset in taxonomy.dataset if dataset.fides_key == fides_key]
        ),
        None,
    )
    return cast(Dataset, dataset)


def evaluate_dataset_reference(
    taxonomy: Taxonomy,
    policy: Policy,
    system: System,
    policy_rule: PolicyRule,
    privacy_declaration: PrivacyDeclaration,
    dataset: Dataset,
) -> List[Violation]:
    """
    Evaluates the constraints of a given rule and dataset that was referenced
    from a given privacy declaration
    """
    evaluation_violation_list = []
    if dataset.data_categories:

        dataset_violation_message = "Declaration ({}) of system ({}) failed rule ({}) from policy ({}) for dataset ({})".format(
            privacy_declaration.name,
            system.fides_key,
            policy_rule.name,
            policy.fides_key,
            dataset.fides_key,
        )

        dataset_result_violations = evaluate_policy_rule(
            taxonomy=taxonomy,
            policy_rule=policy_rule,
            data_subjects=privacy_declaration.data_subjects,
            data_categories=dataset.data_categories,
            data_qualifier=dataset.data_qualifier,
            data_use=privacy_declaration.data_use,
            declaration_violation_message=dataset_violation_message,
        )

        evaluation_violation_list += dataset_result_violations

    for collection in dataset.collections:

        collection_violation_message = "Declaration ({}) of system ({}) failed rule ({}) from policy ({}) for dataset collection ({})".format(
            privacy_declaration.name,
            system.fides_key,
            policy_rule.name,
            policy.fides_key,
            collection.name,
        )

        if collection.data_categories:
            dataset_collection_result_violations = evaluate_policy_rule(
                taxonomy=taxonomy,
                policy_rule=policy_rule,
                data_subjects=privacy_declaration.data_subjects,
                data_categories=collection.data_categories,
                data_qualifier=collection.data_qualifier,
                data_use=privacy_declaration.data_use,
                declaration_violation_message=collection_violation_message,
            )

            evaluation_violation_list += dataset_collection_result_violations

        for field in get_all_level_fields(collection.fields):

            field_violation_message = "Declaration ({}) of system ({}) failed rule ({}) from policy ({}) for dataset field ({})".format(
                privacy_declaration.name,
                system.fides_key,
                policy_rule.name,
                policy.fides_key,
                field.name,
            )

            if field.data_categories:
                field_result_violations = evaluate_policy_rule(
                    taxonomy=taxonomy,
                    policy_rule=policy_rule,
                    data_subjects=privacy_declaration.data_subjects,
                    data_categories=field.data_categories,
                    data_qualifier=field.data_qualifier,
                    data_use=privacy_declaration.data_use,
                    declaration_violation_message=field_violation_message,
                )

                evaluation_violation_list += field_result_violations

    return evaluation_violation_list


def evaluate_privacy_declaration(
    taxonomy: Taxonomy,
    policy: Policy,
    system: System,
    policy_rule: PolicyRule,
    privacy_declaration: PrivacyDeclaration,
) -> List[Violation]:
    """
    Evaluates the contraints of a given rule and privacy declaration. This
    includes additional data set references
    """
    evaluation_violation_list = []

    declaration_violation_message = (
        "Declaration ({}) of system ({}) failed rule ({}) from policy ({})".format(
            privacy_declaration.name,
            system.fides_key,
            policy_rule.name,
            policy.fides_key,
        )
    )

    declaration_result_violations = evaluate_policy_rule(
        taxonomy=taxonomy,
        policy_rule=policy_rule,
        data_subjects=privacy_declaration.data_subjects,
        data_categories=privacy_declaration.data_categories,
        data_qualifier=privacy_declaration.data_qualifier,
        data_use=privacy_declaration.data_use,
        declaration_violation_message=declaration_violation_message,
    )

    evaluation_violation_list += declaration_result_violations

    for dataset_reference in privacy_declaration.dataset_references or []:
        dataset = get_dataset_by_fides_key(
            taxonomy=taxonomy, fides_key=dataset_reference
        )
        if dataset:
            evaluation_violation_list += evaluate_dataset_reference(
                taxonomy=taxonomy,
                policy=policy,
                system=system,
                policy_rule=policy_rule,
                privacy_declaration=privacy_declaration,
                dataset=dataset,
            )
        else:
            echo_red(
                "Dataset ({}) referenced in declaration ({}) could not be found in taxonomy".format(
                    dataset_reference, privacy_declaration.name
                )
            )
            raise SystemExit(1)
    return evaluation_violation_list


def execute_evaluation(taxonomy: Taxonomy) -> Evaluation:
    """
    Check the stated constraints of each Privacy Policy's rules against
    each system's privacy declarations.
    """
    evaluation_violation_list = []
    for policy in taxonomy.policy:
        for rule in policy.rules:
            for system in taxonomy.system:
                for declaration in system.privacy_declarations:

                    evaluation_violation_list += evaluate_privacy_declaration(
                        taxonomy=taxonomy,
                        policy=policy,
                        system=system,
                        policy_rule=rule,
                        privacy_declaration=declaration,
                    )
    status_enum = (
        StatusEnum.FAIL if len(evaluation_violation_list) > 0 else StatusEnum.PASS
    )
    new_uuid = str(uuid.uuid4()).replace("-", "_")
    evaluation = Evaluation(
        fides_key=new_uuid,
        status=status_enum,
        violations=evaluation_violation_list,
    )
    return evaluation


def hydrate_missing_resources(
    url: AnyHttpUrl,
    headers: Dict[str, str],
    missing_resource_keys: List[FidesKey],
    dehydrated_taxonomy: Taxonomy,
) -> Taxonomy:
    """
    Query the server for all of the missing resource keys and
    hydrate a copy of the dehydrated taxonomy with them.
    """

    for resource_name in dehydrated_taxonomy.__fields__:
        server_resources = get_server_resources(
            url=url,
            resource_type=resource_name,
            headers=headers,
            existing_keys=missing_resource_keys,
        )
        setattr(
            dehydrated_taxonomy,
            resource_name,
            getattr(dehydrated_taxonomy, resource_name) + server_resources,
        )
    return dehydrated_taxonomy


def populate_referenced_keys(
    taxonomy: Taxonomy,
    url: AnyHttpUrl,
    headers: Dict[str, str],
    last_keys: List[FidesKey],
) -> Taxonomy:
    """
    Takes in a taxonomy with potentially missing references to fides keys.
    Populates any missing fides_keys recursively and returns the populated taxonomy.

    Recursively calls itself after every hydration to make sure there are no new missing keys.
    """
    missing_resource_keys = get_referenced_missing_keys(taxonomy)
    keys_not_found = set(last_keys).intersection(set(missing_resource_keys))
    if keys_not_found:
        echo_red(f"Missing resource keys: {keys_not_found}")
        raise SystemExit(1)

    if missing_resource_keys:
        taxonomy = hydrate_missing_resources(
            url=url,
            headers=headers,
            missing_resource_keys=missing_resource_keys,
            dehydrated_taxonomy=taxonomy,
        )
        return populate_referenced_keys(
            taxonomy=taxonomy, url=url, headers=headers, last_keys=missing_resource_keys
        )
    return taxonomy


def merge_taxonomies(
    primary_taxonomy: Taxonomy, secondary_taxonomy: Taxonomy
) -> Taxonomy:
    """
    Merges the secondary_taxonomy into the primary_taxonomy while
    preserving all of the existing keys within the primary_taxonomy.
    """
    for resource_name in primary_taxonomy.__fields__:
        # Get the unique set of keys we want to include in the merged taxonomy
        primary_keys = [
            resource.fides_key for resource in getattr(primary_taxonomy, resource_name)
        ]
        secondary_resources = [
            resource
            for resource in getattr(secondary_taxonomy, resource_name)
            if resource.fides_key not in primary_keys
        ]
        # Create a list of all of the resources to go in the merged taxonomy
        resource_list = getattr(primary_taxonomy, resource_name) + secondary_resources
        setattr(primary_taxonomy, resource_name, resource_list)

    return primary_taxonomy


def evaluate(
    url: AnyHttpUrl,
    manifests_dir: str,
    headers: Dict[str, str],
    policy_fides_key: str = "",
    message: str = "",
    local: bool = False,
    dry: bool = False,
) -> Evaluation:
    """
    Perform evaluation for a given Policy. If a policy key is not
    provided, perform an evaluation for all of the Policies in an organzation

    Local Policy definition files will be used as opposed to their
    server-definitions if available.
    """

    # Merge the User-defined taxonomy & Default Taxonomy
    user_taxonomy = parse(manifests_dir)
    taxonomy = merge_taxonomies(user_taxonomy, DEFAULT_TAXONOMY)

    # Determine which Policies will be evaluated
    policies = taxonomy.policy
    if not local:
        # Append server-side Policies if not running in local_mode
        policies = get_evaluation_policies(
            local_policies=policies,
            evaluate_fides_key=policy_fides_key,
            url=url,
            headers=headers,
        )

    validate_policies_exist(policies=policies, evaluate_fides_key=policy_fides_key)
    echo_green(
        "Evaluating the following policies:\n- {}".format(
            "\n- ".join([key.fides_key for key in policies])
        )
    )
    print("-" * 10)

    echo_green("Checking for missing resources...")
    missing_resources = get_referenced_missing_keys(taxonomy)
    if local and missing_resources:
        echo_red(str(missing_resources))
        echo_red("Not all referenced resources exist locally!")
        raise SystemExit(1)
    populate_referenced_keys(taxonomy=taxonomy, url=url, headers=headers, last_keys=[])

    echo_green("Executing Policy evaluation(s)...")
    evaluation = execute_evaluation(taxonomy)
    evaluation.message = message
    if not dry:
        echo_green("Sending the evaluation results to the server...")
        response = api.create(
            url=url,
            resource_type="evaluation",
            json_resource=evaluation.json(exclude_none=True),
            headers=headers,
        )
        handle_cli_response(response, verbose=False)

    if evaluation.status == "FAIL":
        pretty_echo(evaluation.dict(), color="red")
        raise SystemExit(1)
    echo_green("Evaluation passed!")

    return evaluation



"""add mssql (Microsoft SQL Server)

Revision ID: f3841942d90c
Revises: d65e7e921814
Create Date: 2021-12-13 22:15:25.043952

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f3841942d90c"
down_revision = "d65e7e921814"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', 'snowflake', 'redshift', 'mssql')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")


def downgrade():
    op.execute("delete from connectionconfig where connection_type in ('mssql')")
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', 'snowflake', 'redshift')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")



"""
Contains all of the options/arguments used by the CLI commands.
"""

from typing import Callable

import click
from fideslang import model_list


def coverage_threshold_option(command: Callable) -> Callable:
    "Add a flag that assumes yes."
    command = click.option(
        "--coverage-threshold", "-c", type=click.IntRange(0, 100), default=100
    )(command)
    return command


def resource_type_argument(command: Callable) -> Callable:
    "Add the resource_type option."
    command = click.argument(
        "resource_type",
        type=click.Choice(model_list, case_sensitive=False),
    )(command)
    return command


def fides_key_argument(command: Callable) -> Callable:
    "Add the fides_key argument."
    command = click.argument(
        "fides_key",
        type=str,
    )(command)
    return command


def fides_key_option(command: Callable) -> Callable:
    "Add the fides_key option."
    command = click.option(
        "-k",
        "--fides-key",
        default="",
        help="The fides_key of the single policy that you wish to evaluate.",
    )(command)
    return command


def manifests_dir_argument(command: Callable) -> Callable:
    "Add the manifests_dir argument."
    command = click.argument(
        "manifests_dir",
        default=".fides/",
        type=click.Path(exists=True),
    )(command)
    return command


def dry_flag(command: Callable) -> Callable:
    "Add a flag that prevents side-effects."
    command = click.option(
        "--dry", is_flag=True, help="Prevent the persistance of any changes."
    )(command)
    return command


def yes_flag(command: Callable) -> Callable:
    "Add a flag that assumes yes."
    command = click.option(
        "--yes",
        "-y",
        is_flag=True,
        help="Automatically responds 'yes' to any prompts.",
    )(command)
    return command


def verbose_flag(command: Callable) -> Callable:
    "Turns on verbose output."
    command = click.option(
        "--verbose",
        "-v",
        is_flag=True,
        help="Enable verbose output.",
    )(command)
    return command


def include_null_flag(command: Callable) -> Callable:
    "Include null attributes in resource output."
    command = click.option(
        "--include-null",
        is_flag=True,
        help="Includes attributes that would otherwise be null.",
    )(command)
    return command


def organization_fides_key_option(command: Callable) -> Callable:
    "Add the organization_fides_key option."
    command = click.option(
        "--org-key",
        "-k",
        default="default_organization",
        help="The organization_fides_key you wish to export resources for.",
    )(command)
    return command


def output_directory_option(command: Callable) -> Callable:
    "Add the output directory option"
    command = click.option(
        "--output-dir",
        "-d",
        default=".fides/",
        help="The output directory for the data map to be exported to.",
    )(command)
    return command


def credentials_id_option(command: Callable) -> Callable:
    "Use credentials defined within fides config."
    command = click.option(
        "--credentials-id",
        type=str,
        help="Use credentials defined within fides config",
    )(command)
    return command


def connection_string_option(command: Callable) -> Callable:
    "Use connection string option to connect to a database"
    command = click.option(
        "--connection-string",
        type=str,
        help="Use connection string option to connect to a database",
    )(command)
    return command


def okta_org_url_option(command: Callable) -> Callable:
    "Use org url option to connect to okta. Requires options --org-url and --token"
    command = click.option(
        "--org-url",
        type=str,
        help="Use org url option to connect to okta. Requires options --org-url and --token",
    )(command)
    return command


def okta_token_option(command: Callable) -> Callable:
    "Use token option to connect to okta. Requires options --org-url and --token"
    command = click.option(
        "--token",
        type=str,
        help="Use token option to connect to okta. Requires options --org-url and --token",
    )(command)
    return command


def aws_access_key_id_option(command: Callable) -> Callable:
    "Use access key id option to connect to aws. Requires options --access_key_id, access_key and --region"
    command = click.option(
        "--access_key_id",
        type=str,
        help="Use access key id option to connect to aws. Requires options --access_key_id, --secret_access_key and --region",
    )(command)
    return command


def aws_secret_access_key_option(command: Callable) -> Callable:
    "Use access key option to connect to aws. Requires options --access_key_id, --secret_access_key and --region"
    command = click.option(
        "--secret_access_key",
        type=str,
        help="Use access key option to connect to aws. Requires options --access_key_id, --secret_access_key and --region",
    )(command)
    return command


def aws_region_option(command: Callable) -> Callable:
    "Use region option to connect to aws. Requires options --access_key_id, access_key and --region"
    command = click.option(
        "--region",
        type=str,
        help="Use region option to connect to aws. Requires options --access_key_id, --secret_access_key and --region",
    )(command)
    return command


def prompt_username(ctx: click.Context, param: str, value: str) -> str:
    if not value:
        value = click.prompt(text="Username")
    return value


def prompt_password(ctx: click.Context, param: str, value: str) -> str:
    if not value:
        value = click.prompt(text="Password", hide_input=True)
    return value


def username_option(command: Callable) -> Callable:
    command = click.option(
        "-u",
        "--username",
        default="",
        callback=prompt_username,
    )(command)
    return command


def password_option(command: Callable) -> Callable:
    command = click.option(
        "-p",
        "--password",
        default="",
        callback=prompt_password,
    )(command)
    return command


def first_name_option(command: Callable) -> Callable:
    command = click.option(
        "-f",
        "--first-name",
        default="",
    )(command)
    return command


def last_name_option(command: Callable) -> Callable:
    command = click.option(
        "-l",
        "--last-name",
        default="",
    )(command)
    return command



from typing import Optional

from fastapi import Depends, HTTPException
from fastapi.params import Security
from fideslang.validation import FidesKey
from loguru import logger
from sqlalchemy.orm import Session
from starlette.status import (
    HTTP_200_OK,
    HTTP_204_NO_CONTENT,
    HTTP_400_BAD_REQUEST,
    HTTP_404_NOT_FOUND,
    HTTP_422_UNPROCESSABLE_ENTITY,
    HTTP_500_INTERNAL_SERVER_ERROR,
)

from fides.api.ops.api import deps
from fides.api.ops.api.v1.scope_registry import (
    CONNECTION_AUTHORIZE,
    SAAS_CONFIG_CREATE_OR_UPDATE,
    SAAS_CONFIG_DELETE,
    SAAS_CONFIG_READ,
    SAAS_CONNECTION_INSTANTIATE,
)
from fides.api.ops.api.v1.urn_registry import (
    AUTHORIZE,
    CONNECTION_TYPES,
    SAAS_CONFIG,
    SAAS_CONFIG_VALIDATE,
    SAAS_CONNECTOR_FROM_TEMPLATE,
    V1_URL_PREFIX,
)
from fides.api.ops.common_exceptions import FidesopsException
from fides.api.ops.models.connectionconfig import ConnectionConfig, ConnectionType
from fides.api.ops.models.datasetconfig import DatasetConfig
from fides.api.ops.schemas.connection_configuration.connection_config import (
    SaasConnectionTemplateResponse,
    SaasConnectionTemplateValues,
)
from fides.api.ops.schemas.saas.saas_config import (
    SaaSConfig,
    SaaSConfigValidationDetails,
    ValidateSaaSConfigResponse,
)
from fides.api.ops.service.authentication.authentication_strategy import (
    AuthenticationStrategy,
)
from fides.api.ops.service.authentication.authentication_strategy_oauth2_authorization_code import (
    OAuth2AuthorizationCodeAuthenticationStrategy,
)
from fides.api.ops.service.connectors.saas.connector_registry_service import (
    ConnectorRegistry,
    ConnectorTemplate,
    create_connection_config_from_template_no_save,
    load_registry,
    registry_file,
    upsert_dataset_config_from_template,
)
from fides.api.ops.util.api_router import APIRouter
from fides.api.ops.util.connection_util import validate_secrets
from fides.api.ops.util.oauth_util import verify_oauth_client
from fides.lib.exceptions import KeyOrNameAlreadyExists

router = APIRouter(tags=["SaaS Configs"], prefix=V1_URL_PREFIX)

# Helper method to inject the parent ConnectionConfig into these child routes
def _get_saas_connection_config(
    connection_key: FidesKey, db: Session = Depends(deps.get_db)
) -> ConnectionConfig:
    logger.info("Finding connection config with key '{}'", connection_key)
    connection_config = ConnectionConfig.get_by(db, field="key", value=connection_key)
    if not connection_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No connection config with key '{connection_key}'",
        )
    if connection_config.connection_type != ConnectionType.saas:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="This action is only applicable to connection configs of connection type 'saas'",
        )
    return connection_config


def verify_oauth_connection_config(
    connection_config: Optional[ConnectionConfig],
) -> None:
    """
    Verifies that the connection config is present and contains
    the necessary configurations for OAuth2 Authorization Code authentication.
    Returns an HTTPException with the appropriate error message indicating
    which configurations are missing or incorrect.
    """

    if not connection_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail="The connection config cannot be found.",
        )

    saas_config = connection_config.get_saas_config()
    if not saas_config:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail="The connection config does not contain a SaaS config.",
        )

    authentication = saas_config.client_config.authentication
    if not authentication:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail="The connection config does not contain an authentication configuration.",
        )

    if authentication.strategy != OAuth2AuthorizationCodeAuthenticationStrategy.name:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail="The connection config does not use OAuth2 Authorization Code authentication.",
        )


@router.put(
    SAAS_CONFIG_VALIDATE,
    dependencies=[Security(verify_oauth_client, scopes=[SAAS_CONFIG_READ])],
    status_code=HTTP_200_OK,
    response_model=ValidateSaaSConfigResponse,
)
def validate_saas_config(
    saas_config: SaaSConfig,
) -> ValidateSaaSConfigResponse:
    """
    Uses the SaaSConfig Pydantic model to validate the SaaS config
    without attempting to save it to the database.

    Checks that:
    - all required fields are present, all field values are valid types
    - each connector_param only has one of references or identity, not both
    """

    logger.info("Validation successful for SaaS config '{}'", saas_config.fides_key)
    return ValidateSaaSConfigResponse(
        saas_config=saas_config,
        validation_details=SaaSConfigValidationDetails(
            msg="Validation successful",
        ),
    )


@router.patch(
    SAAS_CONFIG,
    dependencies=[Security(verify_oauth_client, scopes=[SAAS_CONFIG_CREATE_OR_UPDATE])],
    status_code=HTTP_200_OK,
    response_model=SaaSConfig,
)
def patch_saas_config(
    saas_config: SaaSConfig,
    db: Session = Depends(deps.get_db),
    connection_config: ConnectionConfig = Depends(_get_saas_connection_config),
) -> SaaSConfig:
    """
    Given a SaaS config element, update the corresponding ConnectionConfig object
    or report failure
    """
    logger.info(
        "Updating SaaS config '{}' on connection config '{}'",
        saas_config.fides_key,
        connection_config.key,
    )
    connection_config.update_saas_config(db, saas_config=saas_config)
    return connection_config.saas_config  # type: ignore


@router.get(
    SAAS_CONFIG,
    dependencies=[Security(verify_oauth_client, scopes=[SAAS_CONFIG_READ])],
    response_model=SaaSConfig,
)
def get_saas_config(
    connection_config: ConnectionConfig = Depends(_get_saas_connection_config),
) -> SaaSConfig:
    """Returns the SaaS config for the given connection config."""

    logger.info("Finding SaaS config for connection '{}'", connection_config.key)
    saas_config = connection_config.saas_config
    if not saas_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No SaaS config found for connection '{connection_config.key}'",
        )
    return saas_config


@router.delete(
    SAAS_CONFIG,
    dependencies=[Security(verify_oauth_client, scopes=[SAAS_CONFIG_DELETE])],
    status_code=HTTP_204_NO_CONTENT,
)
def delete_saas_config(
    db: Session = Depends(deps.get_db),
    connection_config: ConnectionConfig = Depends(_get_saas_connection_config),
) -> None:
    """Removes the SaaS config for the given connection config.
    The corresponding dataset and secrets must be deleted before deleting the SaaS config"""

    logger.info("Finding SaaS config for connection '{}'", connection_config.key)
    saas_config = connection_config.saas_config
    if not saas_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No SaaS config found for connection '{connection_config.key}'",
        )

    fides_key = saas_config.get("fides_key")
    dataset = DatasetConfig.filter(
        db=db,
        conditions=(
            (DatasetConfig.connection_config_id == connection_config.id)
            & (DatasetConfig.fides_key == fides_key)
        ),
    ).first()

    warnings = []

    if not fides_key:
        warnings.append("A fides_key was not found for this SaaS config.")

    if dataset:
        warnings.append(
            f"Must delete the dataset with fides_key '{fides_key}' before deleting this SaaS config."
        )

    # The secrets must be cleared since the SaaS config is used for validation and the secrets
    # might not pass validation once a new SaaS config is added.
    if connection_config.secrets:
        warnings.append(
            "Must clear the secrets from this connection config before deleting the SaaS config."
        )

    if warnings:
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail=" ".join(warnings))

    logger.info("Deleting SaaS config for connection '{}'", connection_config.key)
    connection_config.update(db, data={"saas_config": None})


@router.get(
    AUTHORIZE,
    dependencies=[Security(verify_oauth_client, scopes=[CONNECTION_AUTHORIZE])],
    response_model=str,
)
def authorize_connection(
    db: Session = Depends(deps.get_db),
    connection_config: ConnectionConfig = Depends(_get_saas_connection_config),
) -> Optional[str]:
    """Returns the authorization URL for the SaaS Connector (if available)"""

    verify_oauth_connection_config(connection_config)
    authentication = connection_config.get_saas_config().client_config.authentication  # type: ignore

    try:
        auth_strategy: OAuth2AuthorizationCodeAuthenticationStrategy = AuthenticationStrategy.get_strategy(
            authentication.strategy, authentication.configuration  # type: ignore
        )
        return auth_strategy.get_authorization_url(db, connection_config)
    except FidesopsException as exc:
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail=str(exc))


@router.post(
    SAAS_CONNECTOR_FROM_TEMPLATE,
    dependencies=[Security(verify_oauth_client, scopes=[SAAS_CONNECTION_INSTANTIATE])],
    response_model=SaasConnectionTemplateResponse,
)
def instantiate_connection_from_template(
    saas_connector_type: str,
    template_values: SaasConnectionTemplateValues,
    db: Session = Depends(deps.get_db),
) -> SaasConnectionTemplateResponse:
    """
    Creates a SaaS Connector and a SaaS Dataset from a template.

    Looks up the connector type in the SaaS connector registry and, if all required
    fields are provided, persists the associated connection config and dataset to the database.
    """

    registry: ConnectorRegistry = load_registry(registry_file)
    connector_template: Optional[ConnectorTemplate] = registry.get_connector_template(
        saas_connector_type
    )
    if not connector_template:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"SaaS connector type '{saas_connector_type}' is not yet available in Fidesops. For a list of available SaaS connectors, refer to {CONNECTION_TYPES}.",
        )

    if DatasetConfig.filter(
        db=db,
        conditions=(DatasetConfig.fides_key == template_values.instance_key),  # type: ignore[arg-type]
    ).count():
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"SaaS connector instance key '{template_values.instance_key}' already exists.",
        )

    try:
        connection_config: ConnectionConfig = (
            create_connection_config_from_template_no_save(
                db, connector_template, template_values
            )
        )
    except KeyOrNameAlreadyExists as exc:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=exc.args[0],
        )

    connection_config.secrets = validate_secrets(
        db, template_values.secrets, connection_config
    ).dict()
    connection_config.save(db=db)  # Not persisted to db until secrets are validated

    try:
        dataset_config: DatasetConfig = upsert_dataset_config_from_template(
            db, connection_config, connector_template, template_values
        )
    except Exception:
        connection_config.delete(db)
        raise HTTPException(
            status_code=HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"SaaS Connector could not be created from the '{saas_connector_type}' template at this time.",
        )
    logger.info(
        "SaaS Connector and Dataset {} successfully created from '{}' template.",
        template_values.instance_key,
        saas_connector_type,
    )

    return SaasConnectionTemplateResponse(
        connection=connection_config, dataset=dataset_config.ctl_dataset
    )



from typing import Any, ContextManager, Dict, List, Optional

from celery import Celery, Task
from loguru import logger
from sqlalchemy.orm import Session

from fides.core.config import FidesConfig, get_config
from fides.lib.db.session import get_db_engine, get_db_session

CONFIG = get_config()
MESSAGING_QUEUE_NAME = "fidesops.messaging"


class DatabaseTask(Task):  # pylint: disable=W0223
    _task_engine = None
    _sessionmaker = None

    def get_new_session(self) -> ContextManager[Session]:
        """
        Creates a new Session to be used for each task invocation.

        The new Sessions will reuse a shared `Engine` and `sessionmaker`
        across invocations, so as to reuse db connection resources.
        """
        # only one engine will be instantiated in a given task scope, i.e
        # once per celery process.
        if self._task_engine is None:
            _task_engine = get_db_engine(config=CONFIG)

        # same for the sessionmaker
        if self._sessionmaker is None:
            self._sessionmaker = get_db_session(config=CONFIG, engine=_task_engine)

        # but a new session is instantiated each time the method is invoked
        # to prevent session overlap when requests are executing concurrently
        # when in task_always_eager mode (i.e. without proper workers)
        return self._sessionmaker()


def _create_celery(config: FidesConfig = get_config()) -> Celery:
    """
    Returns a configured version of the Celery application
    """
    logger.info("Creating Celery app...")
    app = Celery(__name__)

    celery_config: Dict[str, Any] = {
        # Defaults for the celery config
        "broker_url": config.redis.connection_url,
        "result_backend": config.redis.connection_url,
        "event_queue_prefix": "fides_worker",
        "task_always_eager": True,
        # Ops requires this to route emails to separate queues
        "task_create_missing_queues": True,
        "task_default_queue": "fides",
    }

    celery_config.update(config.celery)

    app.conf.update(celery_config)

    logger.info("Autodiscovering tasks...")
    app.autodiscover_tasks(
        [
            "fides.api.ops.tasks",
            "fides.api.ops.tasks.scheduled",
            "fides.api.ops.service.privacy_request",
            "fides.api.ops.service.privacy_request.request_runner_service",
        ]
    )
    return app


celery_app = _create_celery()


def get_worker_ids() -> List[Optional[str]]:
    """
    Returns a list of the connected heahtly worker UUIDs.
    """
    try:
        connected_workers = [
            key for key, _ in celery_app.control.inspect().ping().items()
        ]
    except Exception as exception:
        logger.critical(exception)
        connected_workers = []
    return connected_workers



from typing import Any, Dict

from botocore.exceptions import ClientError

from fides.api.ops.schemas.storage.storage import (
    SUPPORTED_STORAGE_SECRETS,
    S3AuthMethod,
    StorageSecrets,
    StorageType,
)
from fides.api.ops.util.storage_authenticator import get_s3_session


def secrets_are_valid(
    secrets: SUPPORTED_STORAGE_SECRETS,
    storage_type: StorageType,
) -> bool:
    """Authenticates upload destination with appropriate upload method"""
    uploader: Any = _get_authenticator_from_config(storage_type)
    return uploader(secrets)


def _s3_authenticator(secrets: Dict[StorageSecrets, Any]) -> bool:
    """Authenticates secrets for s3, returns true if secrets are valid"""
    try:
        get_s3_session(S3AuthMethod.SECRET_KEYS.value, secrets.dict())  # type: ignore
        return True
    except ClientError:
        return False


def _get_authenticator_from_config(storage_type: StorageType) -> Any:
    """Determines which uploader method to use based on storage type"""
    return {
        StorageType.s3.value: _s3_authenticator,
    }[storage_type.value]



"""add column for special_category to data_uses

Revision ID: 7a94527643a1
Revises: 150f234ef1de
Create Date: 2022-03-18 03:22:23.975102

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "7a94527643a1"
down_revision = "150f234ef1de"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("data_uses", sa.Column("special_category", sa.Text(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("data_uses", "special_category")
    # ### end Alembic commands ###
